#ifndef MSP_GL_SL_COMPATIBILITY_H_
#define MSP_GL_SL_COMPATIBILITY_H_

#include <string>
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

class LocationAllocator: private TraversingVisitor
{
private:
	struct Uniform
	{
		int location;
		int desc_set;
		int bind_point;

		Uniform(): location(-1), desc_set(-1), bind_point(-1) { }
	};

	std::map<std::string, std::set<unsigned> > used_locations;
	std::map<std::string, Uniform> uniforms;
	std::map<unsigned, std::set<unsigned> > used_bindings;
	std::vector<VariableDeclaration *> unplaced_variables;
	std::vector<VariableDeclaration *> unbound_textures;
	std::vector<InterfaceBlock *> unbound_blocks;

public:
	void apply(Module &, const Features &);
private:
	void apply(Stage &);

	void allocate_locations(const std::string &);
	void bind_uniform(RefPtr<Layout> &, const std::string &, unsigned);
	void add_layout_value(RefPtr<Layout> &, const std::string &, unsigned);

	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &) { }
};

/** Generates default precision declarations or removes precision declarations
according to the requirements of the target API. */
class PrecisionConverter: private TraversingVisitor
{
private:
	Stage *stage;
	std::set<std::string> have_default;
	NodeList<Statement>::iterator insert_point;
	std::set<Node *> nodes_to_remove;

public:
	PrecisionConverter();

	void apply(Stage &);

private:
	virtual void visit(Block &);
	virtual void visit(Precision &);
	virtual void visit(VariableDeclaration &);
};

/** Converts structures of the syntax tree to match a particular set of
features. */
class LegacyConverter: private TraversingVisitor
{
private:
	Stage *stage;
	Features features;
	VariableDeclaration *frag_out;
	NodeList<Statement>::iterator uniform_insert_point;
	std::set<Node *> nodes_to_remove;

public:
	LegacyConverter();

	virtual void apply(Stage &, const Features &);

private:
	void unsupported(const std::string &);

	virtual void visit(Block &);
	bool check_version(const Version &) const;
	bool check_extension(bool Features::*) const;
	bool supports_stage(Stage::Type) const;
	bool supports_unified_interface_syntax() const;
	virtual void visit(VariableReference &);
	virtual void visit(Assignment &);
	bool supports_unified_sampling_functions() const;
	virtual void visit(FunctionCall &);
	bool supports_interface_layouts() const;
	bool supports_stage_interface_layouts() const;
	bool supports_centroid_sampling() const;
	bool supports_sample_sampling() const;
	bool supports_uniform_location() const;
	bool supports_binding() const;
	virtual void visit(VariableDeclaration &);
	bool supports_interface_blocks(const std::string &) const;
	bool supports_interface_block_location() const;
	virtual void visit(InterfaceBlock &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
