#ifndef MSP_GL_SL_PARSER_H_
#define MSP_GL_SL_PARSER_H_

#include <set>
#include <string>
#include <msp/io/base.h>
#include "preprocessor.h"
#include "syntax.h"
#include "tokenizer.h"

namespace Msp {
namespace GL {
namespace SL {

class Parser
{
private:
	std::string source;
	int base_index;
	int source_index;
	Tokenizer tokenizer;
	Preprocessor preprocessor;
	bool allow_stage_change;
	Module *module;
	Stage *cur_stage;
	std::set<std::string> declared_types;
	std::vector<std::string> errors;

public:
	Parser();
	~Parser();

	Module &parse(const std::string &, const std::string &, int);
	Module &parse(IO::Base &, const std::string &, int);

private:
	void parse_source(const std::string &, int);
	void set_required_version(const Version &);
	void source_reference(unsigned, const std::string &);
	void stage_change(Stage::Type);
	void line_change(int, unsigned);

	std::string expect_type();
	std::string expect_identifier();
	int expect_integer();
	bool check(const std::string &);

	static bool is_interface_qualifier(const std::string &);
	static bool is_sampling_qualifier(const std::string &);
	static bool is_interpolation_qualifier(const std::string &);
	static bool is_precision_qualifier(const std::string &);
	static bool is_qualifier(const std::string &);
	bool is_type(const std::string &);
	bool is_identifier(const std::string &);

	void preprocess();
	void preprocess_version();
	void preprocess_pragma();
	void preprocess_pragma_msp();
	void preprocess_stage();

	template<typename T>
	RefPtr<T> create_node();
	template<typename T>
	RefPtr<T> parse_with_recovery(RefPtr<T> (Parser::*)());
	RefPtr<Statement> parse_global_declaration();
	RefPtr<Statement> parse_statement();
	RefPtr<Import> parse_import();
	RefPtr<Precision> parse_precision();
	RefPtr<Layout> parse_layout();
	template<typename T>
	void parse_block(Block &, bool, RefPtr<T> (Parser::*)());
	RefPtr<Expression> parse_expression(const Operator * = 0);
	RefPtr<Literal> parse_literal();
	RefPtr<BinaryExpression> parse_binary(const RefPtr<Expression> &, const Operator &);
	RefPtr<TernaryExpression> parse_ternary(const RefPtr<Expression> &, const Operator &);
	RefPtr<FunctionCall> parse_function_call(const VariableReference &);
	RefPtr<TypeDeclaration> parse_type_declaration();
	RefPtr<BasicTypeDeclaration> parse_basic_type_declaration();
	RefPtr<ImageTypeDeclaration> parse_image_type_declaration();
	RefPtr<StructDeclaration> parse_struct_declaration();
	RefPtr<VariableDeclaration> parse_variable_declaration();
	RefPtr<VariableDeclaration> parse_variable_declaration_with_layout();
	RefPtr<FunctionDeclaration> parse_function_declaration();
	RefPtr<InterfaceBlock> parse_interface_block();
	RefPtr<Conditional> parse_conditional();
	RefPtr<Iteration> parse_for();
	RefPtr<Iteration> parse_while();
	RefPtr<Passthrough> parse_passthrough();
	RefPtr<Return> parse_return();
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
